# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from pygments.formatters import TerminalFormatter
from pygments.lexers import SqlLexer
import pygments
import sys


def highlight(text):
    if sys.stdout.isatty():  # pragma: nocoverage
        return pygments.highlight(text, SqlLexer(), TerminalFormatter())

    return text


class HighlightStream(object):

    def __init__(self, stream):
        self.stream = stream

    def write(self, chunk):
        self.stream.write(highlight(chunk))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from alchemist.db import session
from alchemist.conf import settings
from importlib import import_module
from alchemist.db.query import Query
from sqlalchemy.orm.util import has_identity
from sqlalchemy.ext.declarative import clsregistry
from sqlalchemy.ext.declarative import declared_attr, DeclarativeMeta, base
import sqlalchemy as sa
import weakref
import re
import six
import warnings


def _component_of(name):
    """Get the root package or module of the passed module.
    """

    # Get the registered package this model belongs to.
    segments = name.split('.')
    while segments:
        # Is this name a registered package?
        test = '.'.join(segments)
        if test in settings.get('COMPONENTS', []):
            # This is the component we are in.
            return test

        # Remove the right-most segment.
        segments.pop()

    if not segments and '.models' in name:
        # No package was found to be registered; attempt to guess the
        # right package name; strip all occurrances of '.models' from the
        # pacakge name.
        return _component_of(name.replace('.models', ''))


#! Global model metadata.
_metadata = sa.MetaData()

#! Componentized model metadata.
_metadata_map = {}

#! Componentized declarative class registries.
_registry_map = {}


def _is_model(name, bases, attrs):

    if name == 'NewBase' and not attrs and not bases:

        # Six contrivance; ensure we skip it.

        attrs['__abstract__'] = True
        return False

    for base in bases:

        if (base.__module__.startswith('alchemist.db.model')
                and base.__name__ == 'NewBase'):

            # This is an internal utility to alchemist.

            continue

        if isinstance(base, ModelBase):

            # This is at least dervied at some point from Model.

            return True

    # This is the base class, Model.

    return False


class ModelBaseProxy(type):

    def __new__(cls, name, bases, attrs):
        # Resolve the metaclass we should be derived from.
        meta_name = settings['MODEL_METACLASS']
        segs = meta_name.split('.')
        meta = getattr(import_module('.'.join(segs[:-1])), segs[-1])

        if not issubclass(bases[0], meta):
            bases = meta,

        return super(ModelBaseProxy, cls).__new__(cls, name, bases, attrs)


class ModelBase(six.with_metaclass(ModelBaseProxy, type)):
    """The metaclass for user-defined models.

    The basic concept is as follows: every model declared in a registered
    component shares the same underlying declarative base. Every model
    declared outside a registered component gets its own declarative base.

    A model is considered declared inside a component whether that be directly
    or nested in a sub-package.
    """

    #! The query class used. The query property returns an instance of this
    #! class.
    __query__ = Query

    #! The component this model belongs to.
    _component = None

    #! A reference to the declarative class registry if the model is
    #! in a registered component.
    __registry = None

    @property
    def query(self):
        """
        Create a session query over this model. ::
            Model.query.all()

        Equivalent to ::
            from alchemist.db import session
            session.query(Model).all()
        """

        return session.query(self)

    @property
    def _decl_class_registry(self):
        return getattr(self, '_ModelBase__registry', None)

    def __new__(cls, name, bases, attrs):

        # Don't process further if this is not a model (ie. the base class).

        if not _is_model(name, bases, attrs):

            return super(ModelBase, cls).__new__(cls, name, bases, attrs)

        # Default `extend_existing`
        if "__table_args__" not in attrs:
            attrs["__table_args__"] = ()
        attrs["__table_args__"] = attrs["__table_args__"] + (
            {"extend_existing": True},)

        # Check if this model is in a registered component and set its
        # declarative registry and metadata to share the global if so; else,
        # generate its own.

        attrs['_component'] = component = _component_of(attrs['__module__'])

        if component:
            if component not in _registry_map:
                _registry_map[component] = weakref.WeakValueDictionary()

            attrs['_ModelBase__registry'] = _registry_map[component]
            attrs['metadata'] = _metadata

        else:
            attrs['_ModelBase__registry'] = weakref.WeakValueDictionary()
            attrs['metadata'] = sa.MetaData()

        return super(ModelBase, cls).__new__(cls, name, bases, attrs)

    def __init__(self, name, bases, attrs):
        # Initialize ourself.
        super(ModelBase, self).__init__(name, bases, attrs)

        # Just short-circuit if we're that six contraption, NewBase.
        if self.__module__ == 'alchemist.db.model' and name == 'NewBase':
            return

        abstract = attrs.get('__abstract__', False)
        if _is_model(name, bases, attrs) and not abstract:

            # Add a reference to the model on the table.

            table = self.__table__
            setattr(table, 'class_', weakref.proxy(self))

            # Add ourself to the componetized metadata.

            if self._component:
                if self._component not in _metadata_map:
                    _metadata_map[self._component] = sa.MetaData()

                _metadata_map[self._component]._add_table(
                    table.name,
                    _metadata.tables[table.name].schema or _metadata.schema,
                    table)


class Model(six.with_metaclass(ModelBase)):

    __abstract__ = True

    __init__ = base._declarative_constructor

    @declared_attr
    def __tablename__(cls):
        """
        Underscorizes the class name combined with the pacakge name in order
        to form a normal name for a table in SQL.
        """
        component = cls._component
        name = '%s.%s' % (component, cls.__name__.lower())
        name = re.sub(r'([A-Z])', r'_\1', name)
        name = re.sub(r'\.+', r'_', name)
        return name

    def save(self, commit=False):
        """Save the changes to the model.

        If the model has not been persisted
        then it adds the model to the declared session. Then it flushes the
        object session and optionally commits it.
        """
        if not has_identity(self):
            # Object has not been persisted to the database.
            session.add(self)

        if commit:
            # Commit the session as requested.
            session.commit()

        else:
            # Just flush the session; do not commit.
            session.flush()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from ... import utils
from .. import metadata, engine
from .._engine import clear_cache
from .._session import refresh
from .utils import HighlightStream
from sqlalchemy_utils import (create_mock_engine, create_database,
                              drop_database, database_exists)
import sys


def op(expression, tables=None, test=None, primary=None, secondary=None,
       names=None, echo=False, commit=True,
       offline=False, verbose=False):

    if verbose:
        url = repr(engine['default'].url)
        utils.print_('*', primary, 'default', url)

    # Offline preparation cannot commit to the database and should always
    # echo output.

    if offline:
        commit = False
        echo = True

    for table in (tables or metadata.sorted_tables):

        if not is_table_included(table, names):
            continue

        # Determine the target engine from the model.

        target = engine['default']

        if not offline and test(target, table):
            continue

        if verbose:
            utils.print_('-', secondary, table.name)

        if echo:
            stream = HighlightStream(sys.stdout)
            mock = create_mock_engine(target, stream)
            expression(mock, table)

        if commit:
            expression(target, table)

    # Refresh the session; ensure any further operations succeed in a
    # new context.

    refresh()


def init(**kwargs):
    """Initialize the specified names in the specified databases.

    The general process is as follows:
      - Ensure the database in question exists
      - Ensure all tables exist in the database.
    """

    # TODO: Iterate through all engines in name set.
    database = kwargs.pop('database', False)
    if database and not database_exists(engine['default'].url):
        create_database(engine['default'].url, encoding='utf8')
        clear_cache()

    expression = lambda target, table: table.create(target)
    test = lambda target, table: table.exists(target)
    op(expression, test=test, primary='init', secondary='create', **kwargs)


def clear(**kwargs):
    """Clear the specified names from the specified databases.

    This can be highly destructive as it destroys tables and when all names
    are removed from a database, the database itself.
    """

    database = kwargs.pop('database', False)
    expression = lambda target, table: table.drop(target)
    test = lambda x, tab: not database_exists(x.url) or not tab.exists(x)

    # TODO: Iterate through all engines in name set.
    if database and database_exists(engine['default'].url):
        drop_database(engine['default'].url)
        clear_cache()

    op(expression, reversed(metadata.sorted_tables), test=test,
       primary='clear', secondary='drop', **kwargs)


def flush(**kwargs):
    """Flush the specified names from the specified databases.

    This can be highly destructive as it destroys all data.
    """

    expression = lambda target, table: target.execute(table.delete())
    test = lambda target, table: not table.exists(target)
    op(expression, reversed(metadata.sorted_tables), test=test,
       primary='flush', secondary='flush', **kwargs)


def is_table_included(table, names):
    """Determines if the table is included by reference in the names.

    A table can be named by its component or its model (using the short-name
    or a full python path).

    eg. 'package.models.SomeModel' or 'package:SomeModel' or 'package'
        would all include 'SomeModel'.
    """

    # No names indicates that every table is included.
    if not names:
        return True

    # Introspect the table and pull out the model and component from it.
    model, component = table.class_, table.class_._component

    # Check for the component name.
    if component in names:
        return True

    # Check for the full python name.
    model_name = '%s.%s' % (model.__module__, model.__name__)
    if model_name in names:
        return True

    # Check for the short name.
    short_name = '%s:%s' % (component, model.__name__)
    if short_name in names:
        return True

    return False



import sqlalchemy as sa
from alchemist import db
from ..models import Entity


class Box(db.Model):

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)


class Plane(Entity):

    id = sa.Column(sa.Integer, sa.ForeignKey(Entity.id), primary_key=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from alchemist import app
import collections
import six

__all__ = [
    'settings',
    'defer'
]


class Settings(collections.Mapping):
    """
    A proxy dictionary-like object that proxies to the current app
    configuration; or, if not in an active application context, pretends
    to be an empty dictionary.
    """

    def __getattr__(self, name):
        try:
            return self[name]

        except KeyError:
            raise AttributeError

    def __getitem__(self, name):
        try:
            return app.application.config[name]

        except (AttributeError, RuntimeError):
            raise KeyError

    def __contains__(self, name):
        try:
            return name in app.application.config

        except (AttributeError, RuntimeError):
            return False

    def __len__(self):
        try:
            return len(app.application.config)

        except (AttributeError, RuntimeError):
            return 0

    def __iter__(self):
        try:
            return iter(app.application.config)

        except:
            return iter(())


# Instantiate lazy-bound settings object.
settings = Settings()


class defer(object):
    """Acts a proxy for a configuration setting.
    """

    def __init__(self, expression):
        self._expression = expression

    def resolve(self):
        from . import settings
        locals().update(settings)
        six.exec_('value = (%s)' % self._expression, locals(), globals())
        return value  # noqa



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division


class ImproperlyConfigured(BaseException):
    """A required configuration option is missing or mis-configured.
    """



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
import alchemist
import contextlib
import py
from alchemist.test import settings
from flask import Flask
from pytest import raises
from sqlalchemy.engine.result import ResultProxy
from sqlalchemy.exc import OperationalError
from . import utils


try:   # pragma: nocoverage
    from unittest import mock

except ImportError:  # pragma: nocoverage
    import mock


class TestEngine:

    @staticmethod
    def _clear_cache():
        from alchemist.db._engine import Engine
        Engine.__getitem__._cache.clear()

    def setup(self):
        utils.unload_modules('alchemist')
        self._clear_cache()
        self.app = Flask('alchemist')
        self.context = self.app.app_context()
        self.context.push()

    def teardown(self):
        self.context.pop()
        self._clear_cache()

    def test_default_repr(self):
        from alchemist import db

        uri = 'sqlite:///:memory:'
        with settings(self.app, DATABASES={'default': uri}):

            assert repr(db.engine) == repr(db.engine['default'])

    def test_default(self):
        from alchemist import db

        uri = 'sqlite:///:memory:'
        with settings(self.app, DATABASES={'default': uri}):

            assert db.engine.url.drivername == 'sqlite'
            assert db.engine.url.database == ':memory:'

    def test_alternative(self):
        from alchemist import db

        uri = 'sqlite:///:memory:'
        with settings(self.app, DATABASES={'other': uri}):

            assert db.engine['other'].url.drivername == 'sqlite'
            assert db.engine['other'].url.database == ':memory:'

    def test_missing(self):
        from alchemist import db, exceptions

        with settings(self.app):

            with raises(exceptions.ImproperlyConfigured):
                db.engine['default']

    def test_missing_database(self):
        from alchemist import db, exceptions

        uri = 'sqlite:///:memory:'
        with settings(self.app, DATABASES={'other': uri}):

            with raises(exceptions.ImproperlyConfigured):
                db.engine['default']

    def test_usage(self):
        from alchemist import db

        uri = 'sqlite:///:memory:'
        with settings(self.app, DATABASES={'default': uri}):

            with contextlib.closing(db.engine.connect()) as connection:

                result = connection.execute("BEGIN")
                assert isinstance(result, ResultProxy)

    def test_lookup(self):
        from alchemist import db

        uri = 'sqlite:///:memory:'
        with settings(self.app, DATABASES={'default': uri}):

            assert 'connect' in dir(db.engine)

    def test_expanded(self):
        from alchemist import db

        config = {
            'engine': 'sqlite',
            'name': 'name',
        }

        with settings(self.app, DATABASES={'default': config}):

            assert db.engine.url.drivername == 'sqlite'


class TestSession:

    def setup(self):
        utils.unload_modules('alchemist')
        self.app = Flask('alchemist')

    def test_unbound(self):
        from alchemist import db

        assert not db.session

    def test_acquire_no_database(self):
        from alchemist import db, exceptions

        with raises(exceptions.ImproperlyConfigured), self.app.app_context():
            assert db.session

    def test_acquire_default(self):
        from alchemist import db

        config = {'default': 'sqlite:///:memory:'}
        with settings(self.app, DATABASES=config), self.app.app_context():
            assert db.session

    def test_repr(self):
        from alchemist import db

        config = {'default': 'sqlite:///:memory:'}
        with settings(self.app, DATABASES=config), self.app.app_context():
            text = '<Session(bind=%r)>' % db.engine['default']
            assert repr(db.session) == text


class TestModel:

    def setup(self):
        utils.unload_modules('alchemist')

        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        self.context.push()

        alchemist.configure(self.app)

    def teardown(self):
        self.context.pop()

    def test_metadata(self):
        from alchemist import db

        assert 'alchemist_tests_a_entity' in db.metadata.tables
        assert 'alchemist_tests_a_box' in db.metadata.tables
        assert 'alchemist_tests_a_b_box' in db.metadata.tables
        assert 'alchemist_tests_a_b_plane' in db.metadata.tables

    def test_registry(self):
        from alchemist import db

        assert 'Entity' in db.registry['alchemist.tests.a']
        assert 'Box' in db.registry['alchemist.tests.a']
        assert 'Box' in db.registry['alchemist.tests.a.b']
        assert 'Plane' in db.registry['alchemist.tests.a.b']


class TestQuery:

    def setup(self):
        utils.unload_modules('alchemist')

        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        self.context.push()

        alchemist.configure(self.app)

    def teardown(self):
        self.context.pop()

    def test_query(self):
        from alchemist import db
        from .a.models import Entity

        assert isinstance(db.session.query(Entity), db.Query)
        assert isinstance(Entity.query, db.Query)


class TestInitializeOperation:

    def setup(self):
        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        self.context.push()

        alchemist.configure(self.app)

    def teardown(self):
        from alchemist import db
        db.clear()

        self.context.pop()

    def test_init(self):
        from alchemist import db
        from .a.models import Entity

        with raises(OperationalError):
            assert Entity.query.all()

        db.init()

        assert Entity.query.all() == []

    def test_init_component(self):
        from alchemist import db
        from .a.models import Entity

        with raises(OperationalError):
            assert Entity.query.all()

        db.init(names=['alchemist.tests.a'])

        assert Entity.query.all() == []

    def test_init_full_name(self):
        from alchemist import db
        from .a.models import Entity

        with raises(OperationalError):
            assert Entity.query.all()

        db.init(names=['alchemist.tests.a.models.Entity'])

        assert Entity.query.all() == []

    def test_init_short_name(self):
        from alchemist import db
        from .a.models import Entity

        with raises(OperationalError):
            assert Entity.query.all()

        db.init(names=['alchemist.tests.a:Entity'])

        assert Entity.query.all() == []

    def test_init_twice(self):
        from alchemist import db
        from .a.models import Entity

        with raises(OperationalError):
            assert Entity.query.all()

        db.init(names=['alchemist.tests.a:Entity'])

        assert Entity.query.all() == []

        db.init(names=['alchemist.tests.a:Entity'])

        assert Entity.query.all() == []

    def test_init_verbose(self):
        from alchemist import db

        capture = py.io.StdCapture(out=True, in_=False)
        db.init(verbose=True)
        out, err = capture.done()

        text = err.read()

        assert 'create' in text
        assert 'alchemist_tests_a_entity' in text

    def test_init_offline(self):
        from alchemist import db
        from .a.models import Entity

        with raises(OperationalError):
            assert Entity.query.all()

        capture = py.io.StdCapture(out=True, in_=False)
        db.init(offline=True)
        _, _ = capture.done()

        with raises(OperationalError):
            assert Entity.query.all()

    def test_init_echo(self):
        from alchemist import db

        capture = py.io.StdCapture(out=True, in_=False)
        db.init(echo=True)
        out, err = capture.done()

        text = out.read()

        assert 'CREATE TABLE' in text
        assert 'alchemist_tests_a_entity' in text


class TestClearOperation:

    def setup(self):
        from alchemist import db

        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        self.context.push()

        db.init()

    def teardown(self):
        from alchemist import db

        db.clear()

        self.context.pop()

    def test_clear(self):
        from alchemist import db
        from .a.models import Entity

        assert Entity.query.all() == []

        db.clear()

        with raises(OperationalError):
            assert Entity.query.all()

    def test_clear_name(self):
        from alchemist import db
        from .a.models import Entity

        assert Entity.query.all() == []

        db.clear(names=['alchemist.tests.a'])

        with raises(OperationalError):
            assert Entity.query.all()

    def test_clear_twice(self):
        from alchemist import db
        from .a.models import Entity

        assert Entity.query.all() == []

        db.clear()

        with raises(OperationalError):
            assert Entity.query.all()

        db.clear()

        with raises(OperationalError):
            assert Entity.query.all()

    def test_clear_verbose(self):
        from alchemist import db

        capture = py.io.StdCapture(out=True, in_=False)
        db.clear(verbose=True)
        out, err = capture.done()

        text = err.read()

        assert 'drop' in text
        assert 'alchemist_tests_a_entity' in text

    def test_clear_offline(self):
        from alchemist import db
        from .a.models import Entity

        assert Entity.query.all() == []

        capture = py.io.StdCapture(out=True, in_=False)
        db.clear(offline=True)
        _, _ = capture.done()

        assert Entity.query.all() == []

    def test_clear_echo(self):
        from alchemist import db

        capture = py.io.StdCapture(out=True, in_=False)
        db.clear(echo=True)
        out, err = capture.done()

        text = out.read()

        assert 'DROP TABLE' in text
        assert 'alchemist_tests_a_entity' in text


class TestFlushOperation:

    def setup(self):
        from alchemist import db

        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        self.context.push()

        db.init()

        from .a.models import Entity

        db.session.add(Entity())
        db.session.commit()

    def teardown(self):
        from alchemist import db

        db.clear()

        self.context.pop()

    def test_flush(self):
        from alchemist import db
        from .a.models import Entity

        assert len(Entity.query.all()) == 1

        db.flush()

        assert Entity.query.all() == []

    def test_flush_non_existing(self):
        from alchemist import db
        from .a.models import Entity

        assert len(Entity.query.all()) == 1

        db.clear()
        db.flush()

        with raises(OperationalError):
            assert Entity.query.all()

    def test_flush_name(self):
        from alchemist import db
        from .a.models import Entity

        assert len(Entity.query.all()) == 1

        db.flush(names=['alchemist.tests.a'])

        assert Entity.query.all() == []

    def test_flush_twice(self):
        from alchemist import db
        from .a.models import Entity

        assert len(Entity.query.all()) == 1

        db.flush()

        assert Entity.query.all() == []

        db.flush()

        assert Entity.query.all() == []

    def test_flush_verbose(self):
        from alchemist import db

        capture = py.io.StdCapture(out=True, in_=False)
        db.flush(verbose=True)
        out, err = capture.done()

        text = err.read()

        assert 'flush' in text
        assert 'alchemist_tests_a_entity' in text

    def test_flush_offline(self):
        from alchemist import db
        from .a.models import Entity

        assert len(Entity.query.all()) == 1

        capture = py.io.StdCapture(out=True, in_=False)
        db.flush(offline=True)
        _, _ = capture.done()

        assert len(Entity.query.all()) == 1

    def test_flush_echo(self):
        from alchemist import db

        capture = py.io.StdCapture(out=True, in_=False)
        db.flush(echo=True)
        out, err = capture.done()

        text = out.read()

        assert 'DELETE FROM' in text
        assert 'alchemist_tests_a_entity' in text


class TestShellOperation:

    @staticmethod
    def _clear_cache():
        from alchemist.db._engine import Engine
        Engine.__getitem__._cache.clear()

    def setup(self):
        self._clear_cache()
        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        self.context.push()

    def teardown(self):
        self.context.pop()
        self._clear_cache()

    def test_shell_sqlite3(self):
        from alchemist import db

        with settings(self.app, DATABASES={'default': 'sqlite:///:memory:'}):

            target = mock.MagicMock()
            with mock.patch('os.execvp', target):
                db.shell()

            target.assert_called_with('sqlite3', ['sqlite3', ':memory:'])

    def test_shell_mysql(self):
        from alchemist import db

        uri = 'mysql+oursql://user:pass@host:55/name'
        with settings(self.app, DATABASES={'default': uri}):

            target = mock.MagicMock()
            with mock.patch('os.execvp', target):
                db.shell()

            target.assert_called_with('mysql', [
                'mysql', '--user=user', '--password=pass',
                '--host=host', '--port=55', 'name'])

    def test_shell_mysql_socket(self):
        from alchemist import db

        uri = ('mysql+oursql://user:pass@localhost/name?'
               'unix_socket=/some/file/over/there.socket')
        with settings(self.app, DATABASES={'default': uri}):

            target = mock.MagicMock()
            with mock.patch('os.execvp', target):
                db.shell()

            target.assert_called_with('mysql', [
                'mysql', '--user=user', '--password=pass',
                '--socket=/some/file/over/there.socket', 'name'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from flask.ext.script import Command, Option
from alchemist import db


common_options = [
    Option(dest='names', nargs='*',
           help='The names of components or models to limit the scope to.'),
    Option('--echo', action='store_true', required=False, default=False,
           help='Print the SQL statements as they are processed to stdout.'),
    Option('--dry-run', action='store_false', dest='commit',
           required=False, default=True,
           help='Do not commit or finalize any operation.'),
    Option('--offline', action='store_true', required=False, default=False,
           help='Make no connection to the database server.'),
    Option('--quiet', '-q', dest='verbose',
           action='store_false', required=False, default=True),
]


class Initialize(Command):
    """Initialize the database; create all binds and tables.
    """

    name = 'init'

    namespace = 'db'

    option_list = common_options + [
        Option('--database', dest='database',
               action='store_true', required=False, default=False,
               help='Initialize the database before tables.'),
    ]

    def run(self, *args, **kwargs):
        db.init(*args, **kwargs)


class Clear(Command):
    """Clear the database; remove all binds and tables.
    """

    name = 'clear'

    namespace = 'db'

    option_list = common_options + [
        Option('--database', dest='database',
               action='store_true', required=False, default=False,
               help='Clear the database after tables.'),
    ]

    def run(self, *args, **kwargs):
        db.clear(*args, **kwargs)


class Flush(Command):
    """Flush the database; remove all data.
    """

    name = 'flush'

    namespace = 'db'

    option_list = common_options

    def run(self, *args, **kwargs):
        db.flush(*args, **kwargs)


class Shell(Command):
    """Runs the CLI for the specified database.
    """

    name = 'shell'

    namespace = 'db'

    option_list = [
        Option(dest='database', nargs='?', default='default',
               help='The database to drop into the shell for.'),
    ]

    def run(self, *args, **kwargs):
        db.shell(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from .server import Server
from .shell import Shell
from .settings import Settings
from .load import Load
from .db import (Initialize, Clear, Flush, Shell as DBShell)
from .test import Test

__all__ = [
    'Server',
    'Shell',
    'Initialize', 'Clear', 'Flush', 'DBShell',
    'Settings',
    'Load',
    'Test'
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from flask.ext import script, components
import colorama
import sys


class Manager(script.Manager):

    def __init__(self, app, **kwargs):

        # Initialize color output.
        if sys.stdout.isatty():  # pragma: nocoverage
            colorama.init()

        # Disable loading of default flask-script commands.
        kwargs.setdefault('with_default_commands', False)

        super(Manager, self).__init__(app, **kwargs)

        # Discover commands using the flask-components utility.
        for component in components.find('commands', app):
            for command in component.values():
                if (command and isinstance(command, type) and
                        issubclass(command, (script.Command, script.Manager))):
                    self.add_command(command)

    def __call__(self, app=None, **kwargs):
        # Just ignore these features of flask-script as we wrap it and return
        # our configured application.
        return app or self.app

    def run(self, commands=None, default_command=None, context=None):
        """
            Context: A dict of namespaces as the key, and their objects as the
            value. Used to easily inject code into the shell's runtime env.
        """

        if commands:
            self._commands.update(commands)

            # HACK: Overriding the old shell isn't cool.
            # Should do it by default.
        from alchemist.commands import Shell
        self._commands['shell'] = Shell(context=context)

        if default_command is not None and len(sys.argv) == 1:
            sys.argv.append(default_command)

        try:
            result = self.handle(sys.argv[0], sys.argv[1:])
        except SystemExit as e:
            result = e.code

        sys.exit(result or 0)

# Monkey path flask-script (until it can better handle normal WSGI
# applications)
script.Manager.__call__ = Manager.__call__



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from contextlib import contextmanager
from alchemist import db
import requests
import json


@contextmanager
def settings(app, **kwargs):
    """Save current application config and restore after block.

    Allows override with the keyword arguments.
    """

    # Save the current application config to restore.
    _config = dict(app.config)

    # Make requested changes for this scope.
    app.config.update(kwargs)

    yield

    # Restore the previous application config.
    app.config = _config


class TestBase:

    #! Host at which the test server will be intercepted at.
    host = 'localhost'

    #! Port at which the test server will be intercepted at.
    port = 5000

    def setup_class(cls):
        # Create a shortcut for querying because we're all lazy and we
        # know it.
        cls.Q = lambda s, x: db.session.query(x)

    def teardown(self):
        # Rollback the session.
        db.session.rollback()

        # Flush the database access layer.
        db.flush()

        # Commit all of the deletes.
        db.session.commit()

        # TODO: Re-load any desired fixtures.

    def request(self, path='', url=None, data=None, format='json',
                *args, **kwargs):
        # Helper to forward the port and host if url is not specified.
        if url is None:
            url = 'http://{}:{}'.format(self.host, self.port)

        # Get request header dictionary.
        kwargs.setdefault('headers', {})
        headers = kwargs.get('headers')

        # Automatically serialize and send data.
        if format == 'json':
            if data is None:
                data = {}

            kwargs['data'] = json.dumps(data)
            headers.setdefault('Content-Type', 'application/json')

        else:
            kwargs['data'] = data

        # Set some defaults for requests.
        kwargs.setdefault('allow_redirects', False)

        # Forward to requests.
        response = requests.request(*args, url=url + path, **kwargs)

        # Return the response.
        return response

    def head(self, *args, **kwargs):
        kwargs.setdefault('method', 'HEAD')
        return self.request(*args, **kwargs)

    def options(self, *args, **kwargs):
        kwargs.setdefault('method', 'OPTIONS')
        return self.request(*args, **kwargs)

    def get(self, *args, **kwargs):
        kwargs.setdefault('method', 'GET')
        return self.request(*args, **kwargs)

    def post(self, *args, **kwargs):
        kwargs.setdefault('method', 'POST')
        return self.request(*args, **kwargs)

    def put(self, *args, **kwargs):
        kwargs.setdefault('method', 'PUT')
        return self.request(*args, **kwargs)

    def patch(self, *args, **kwargs):
        kwargs.setdefault('method', 'PATCH')
        return self.request(*args, **kwargs)

    def delete(self, *args, **kwargs):
        kwargs.setdefault('method', 'DELETE')
        return self.request(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from flask.ext import script
import sys
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer
import pygments
import pprint


class Settings(script.Command):
    """Show the collected configuration settings.
    """

    name = 'settings'

    def __call__(self, app, **kwargs):

        text = pprint.pformat(dict(app.config))

        if sys.stdout.isatty():  # pragma: nocoverage
            text = pygments.highlight(
                text, PythonLexer(), TerminalFormatter())

        print(text)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
import types
import sys
import six
import functools
from termcolor import colored
from werkzeug.wsgi import DispatcherMiddleware as BaseDispatcherMiddleware
from importlib import import_module


def make_module_class(name):
    """Takes the module referenced by name and make it a full class.
    """

    source = sys.modules[name]

    members = vars(source)
    is_descriptor = lambda x: not isinstance(x, type) and hasattr(x, '__get__')
    descriptors = {k: v for (k, v) in members.items() if is_descriptor(v)}
    members = {k: v for (k, v) in members.items() if k not in descriptors}
    descriptors['__source'] = source

    target = type(name, (types.ModuleType,), descriptors)(name)
    target.__dict__.update(members)

    sys.modules[name] = target


def memoize(obj):
    cache = obj._cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = obj(*args, **kwargs)

        return cache[args]

    return memoizer


def print_(indicator, name, target='', extra=''):
    six.print_(
        colored(' ' + indicator, 'white', attrs=['dark']),
        colored(name, 'cyan'),
        colored(target, 'white'),
        colored(extra, 'white', attrs=['dark']),
        file=sys.stderr)


class DispatcherMiddleware(BaseDispatcherMiddleware):

    def __init__(self, app, mounts):
        # Load each mount (if needed) using import_module
        for route, application in mounts.items():
            if isinstance(application, six.string_types):
                module, name = application.rsplit(".", 1)
                mounts[route] = getattr(import_module(module), name)

        # Continue initialization.
        super().__init__(app, mounts)

    def __getattr__(self, name):
        return getattr(self.app, name)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from flask.ext import script
from werkzeug.serving import run_simple


class Server(script.Server):

    name = 'run'

    option_list = script.Server.option_list + (
        script.Option('--mode', '-m', dest='mode',
                      action='store', required=False, default="werkzeug"),
    )

    def __init__(self, *args, **kwargs):

        # Set all defaults to None as they will be defaulted in handle
        # when we have access to an application.
        kwargs.setdefault('port', None)
        kwargs.setdefault('host', None)
        kwargs.setdefault('use_debugger', None)
        kwargs.setdefault('use_reloader', None)
        kwargs.setdefault('threaded', None)
        kwargs.setdefault('processes', None)

        super(Server, self).__init__(*args, **kwargs)

    def __call__(self, app, **kwargs):

        # Collect default server configuration from the application config.
        config = app.config
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        kwargs.setdefault('host', config.get('SERVER_HOST', "127.0.0.1"))
        kwargs.setdefault('port', config.get('SERVER_PORT', 5000))
        kwargs.setdefault('use_debugger', config.get('SERVER_DEBUG'))
        kwargs.setdefault('use_reloader', config.get('SERVER_RELOAD'))
        kwargs.setdefault('threaded', config.get('SERVER_THREADED'))
        kwargs.setdefault('processes', config.get('SERVER_PROCESSES', 1))

        # Spin up the server.
        mode = kwargs.get("mode", "werkzeug")
        if mode == "werkzeug":
            processes = kwargs["processes"]
            run_simple(kwargs["host"],
                       kwargs["port"],
                       app,
                       use_debugger=kwargs["use_debugger"],
                       use_reloader=kwargs["use_reloader"],
                       threaded=kwargs["threaded"],
                       processes=(False if processes == 1 else processes))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
import sys


def unload_modules(name):
    """Unload all modules where the name begins with the passed name.
    """

    for key in list(sys.modules.keys()):

        if key.startswith(name):

            del sys.modules[key]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from . import utils
from flask.ext import components
from importlib import import_module
import flask
import inspect
import os
import sys
import sqlalchemy as sa
from sqlalchemy_utils import coercion_listener


def configure(self, app):

    # Detect if we are being invoked by a test runner.
    # Checks the first and second arguments to determine if we are being
    # run by a test runner.
    # Able to match `py.test`, `nosetests`, `alchemist test`,
    # and `python setup.py test`.

    app.config['TESTING'] = False
    count = 3 if os.path.basename(sys.argv[0]).startswith('python') else 2
    for argument in sys.argv[0:count]:
        if 'test' in argument:
            # We are, in fact, being run by a test runner.
            app.config['TESTING'] = True
            break

    # Gather configuration from the following places:
    #  1. alchemist.conf.default_settings
    app.config.from_object('alchemist.conf.default_settings')

    #  2. <package>.settings

    try:
        app.config.from_object('%s.settings' % app.name)

    except ImportError:
        pass

    #  3. Gather configuration from each registered component.

    name = app.name
    for component in components.find('settings', app, raw=True):
        if component.__package__ != name or component.__name__ != name:
            app.config.from_object(component)

    #  4. $ALCHEMIST_SETTINGS_MODULE

    app.config.from_envvar('ALCHEMIST_SETTINGS_MODULE', silent=True)

    #  5. $<package>_SETTINGS_MODULE

    var = '%s_SETTINGS_MODULE' % app.name.replace('.', '_').upper()
    app.config.from_envvar(var, silent=True)

    # Resolve deferred configuration.

    from alchemist.conf import defer

    with app.app_context():
        for name, value in app.config.items():
            if isinstance(value, defer):
                app.config[name] = value.resolve()

    # Gather and import all models modules or packages of the
    # registered components.

    components.find('models', app, raw=True)

    # Register the coercion listener.
    sa.event.listen(sa.orm.mapper, 'mapper_configured', coercion_listener)


@utils.memoize
def _get_application_from_name(self, name):

    # Take the passed name (eg. 'sandbox') and check the following
    # locations for a flask application:
    #   <name>.app
    #   <name>.application
    #   <name>.app.application
    #   <name>.app.app

    if not name:
        return None

    for part in ('', '.app'):
        try:
            module = import_module(name + part)

        except ImportError:
            continue

        app = getattr(module, 'app', None)
        if app and isinstance(app, flask.Flask):
            return app

        app = getattr(module, 'application', None)
        if app and isinstance(app, flask.Flask):
            return app

    return self._get_application_from_name('.'.join(name.split('.')[:-1]))


@utils.memoize
def _find_application(self):

    # Check for an environment variable declaring where the central
    # application lives.

    name = os.environ.get('ALCHEMIST_APPLICATION')
    if name:
        app = self._get_application_from_name(name)
        if app:
            return app

    # Inspect the stack to find the application.
    # This is intended for ease during development and for production
    # to lock-in the application via the environment variable.

    for frame in reversed(inspect.stack()[1:]):
        name = frame[0].f_globals.get('__package__')
        if (name and (not name.startswith('alchemist')
                      or name.startswith('alchemist.tests.a'))):
            app = self._get_application_from_name(name)
            if app:
                return app


@property
def application(self):

    # Short-circuit this if we're in an application context.

    if flask.current_app:
        return flask.current_app

    # Attempt to find the application.

    return self._find_application()


utils.make_module_class(__name__)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from . import engine, query
from flask import appcontext_tearing_down, g
from sqlalchemy import orm
from werkzeug import LocalProxy


class Session(orm.Session):

    def __init__(self, **kwargs):

        # Default the bind and query class.

        # TODO: Serisouly consider using autocommit.
        #http://docs.sqlalchemy.org/en/rel_0_8/orm/session.html#autocommit-mode

        kwargs.setdefault('autocommit', False)
        kwargs.setdefault('autoflush', True)
        kwargs.setdefault('query_cls', query.Query)

        # Bind to the default database engine for now.
        # Once we have db.Model built we can ask that what to bind to
        # upon operation.

        # NOTE: We sub into engine instead of just letting it be because
        # we want to raise as soon as possible if the user forgot their
        # config.

        kwargs.setdefault('bind', engine['default'])

        super(Session, self).__init__(**kwargs)

    def __repr__(self):
        return '<Session(bind=%r)>' % self.bind


def _get_session():
    _session = getattr(g, '_session', None)
    if _session is None:
        _session = g._session = Session()

    return _session


@appcontext_tearing_down.connect
def _teardown_session(*args, **kwargs):
    """
    Ensure the session is closd on the application context teardown which
    happens on the end of a request or interactive session.

    This ensures that the database connnection is closed appropriately.
    """

    _session = getattr(g, '_session', None)
    if _session is not None:
        _session.close()


session = LocalProxy(_get_session)


def refresh():
    if session:
        session.expire_all()
        session.expunge_all()
        session.commit()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from alchemist import utils, exceptions
from alchemist.conf import settings
import sqlalchemy as sa
import six
import threading
from sqlalchemy.engine.url import URL, make_url


class Engine(object):

    def __getattribute__(self, name):
        return getattr(self['default'], name)

    def __dir__(self):
        return dir(self['default'])

    def __repr__(self):
        return repr(self['default'])

    @utils.memoize
    def __getitem__(self, name):

        if 'DATABASES' not in settings:
            raise exceptions.ImproperlyConfigured(
                'DATABASES not configured in project settings.')

        if name not in settings['DATABASES']:
            raise exceptions.ImproperlyConfigured(
                '%r not present in DATABASES configuration.' % name)

        config = settings['DATABASES'][name]

        if isinstance(config, six.string_types):
            url = make_url(config)
            options = {}

        else:
            config = dict(map(lambda i: (i[0].lower(), i[1]), config.items()))
            options = config.get('options', {})
            url = URL(
                config['engine'],
                username=config.get('username', config.get('user')),
                password=config.get('password', config.get('pass')),
                host=config.get('hostname', config.get('host')),
                port=config.get('port'),
                database=config.get('name', config.get('database')))

        # If alchemist is invoked by a test runner we should switch to using
        # testing databases.

        if settings['TESTING']:

            if url.drivername.startswith('sqlite'):

                # Switch to using an in-memory database for sqlite.
                url.database = ':memory:'

            else:

                # Switch to using a named testing database for other dialects.
                ident = threading.current_thread().ident
                url.database = 'test_%s_%s' % (url.database, ident)

        # Apply MySQL hacks to make MySQL play nice.
        pool_size = None
        pool_recycle = None
        if url.drivername.startswith('mysql'):
            pool_size = 10
            pool_recycle = 7200

        # Get "global" options for the database engine.
        pool_size = settings.get('DATABASE_POOL_SIZE', pool_size)
        if pool_size:
            options.setdefault('pool_size', pool_size)

        pool_recycle = settings.get('DATABASE_POOL_RECYCLE', pool_recycle)
        if pool_recycle:
            options.setdefault('pool_recycle', pool_recycle)

        pool_timeout = settings.get('DATABASE_POOL_TIMEOUT')
        if pool_timeout:
            options.setdefault('pool_timeout', pool_timeout)

        # Forward configuration to sqlalchemy and create the engine.
        engine = sa.create_engine(url, **options)

        if settings["DEBUG"]:
            # Create a no-op listener if we're in debug mode.
            from sqlalchemy.event import listen
            listen(engine, "after_cursor_execute", lambda *a, **kw: None)

        # Return the created engine.
        return engine


def clear_cache():
    Engine.__getitem__._cache.clear()


engine = Engine()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division
from alchemist import test
import alchemist
from flask import Flask
import sys
import io
import py
from . import utils


try:   # pragma: nocoverage
    from unittest import mock

except ImportError:  # pragma: nocoverage
    import mock


class CommandTest:

    def setup(self):
        self.app = Flask('alchemist')
        self.app.config['COMPONENTS'] = ['alchemist']

        alchemist.configure(self.app)

    def _run(self, command, patch=None):
        from alchemist import management

        if patch is None:
            patch = self.patch

        target = mock.MagicMock()
        with mock.patch(patch, target):
            manager = management.Manager(self.app)

            _argv = sys.argv
            sys.argv = ['alchemist'] + command
            try:
                manager.run()

            except SystemExit:
                return target

            finally:
                sys.argv = _argv


class TestServer(CommandTest):

    patch = 'flask.ext.script.Server.handle'

    def test_default(self):
        target = self._run(['run'])
        target.assert_called_with(
            self.app, host='::1', port=8000, use_debugger=True,
            threaded=False, processes=1,
            passthrough_errors=False, use_reloader=True)

    def test_override(self):
        config = {
            'SERVER_HOST': 'app.me',
            'SERVER_PORT': 5000,
            'SERVER_RELOAD': False,
            'SERVER_PROCESSES': 10,
            'SERVER_THREADED': True,
            'SERVER_DEBUG': False
        }

        with test.settings(self.app, **config):
            target = self._run(['run'])
            target.assert_called_with(
                self.app, passthrough_errors=False,
                host=config['SERVER_HOST'],
                port=config['SERVER_PORT'],
                use_reloader=config['SERVER_RELOAD'],
                use_debugger=config['SERVER_DEBUG'],
                processes=config['SERVER_PROCESSES'],
                threaded=config['SERVER_THREADED'])


class TestShell(CommandTest):

    patch = 'flask.ext.script.Shell.run'

    def test_run(self):
        target = self._run(['shell'])
        target.assert_called()

    def test_context(self):
        utils.unload_modules('alchemist')

        from alchemist.commands import Shell
        shell = Shell()

        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()
        with self.context:

            alchemist.configure(self.app)

            context = shell.make_context()

            assert 'db' in context
            assert 'session' in context
            assert 'Entity' in context

    def test_pipe(self):
        utils.unload_modules('alchemist')

        self.app = Flask('alchemist.tests.a')
        self.context = self.app.app_context()

        with self.context:
            alchemist.configure(self.app)

            stdin = sys.stdin
            sys.stdin = io.StringIO('print(Entity)\nprint(session)')

            capture = py.io.StdCapture(out=True, in_=False)
            self._run(['shell', '--pipe'])
            out, err = capture.done()

            text = ("<class 'alchemist.tests.a.models.Entity'>\n"
                    "<Session(bind=Engine(sqlite:///:memory:))>\n")

            assert text == out.read()

            sys.stdin = stdin


class TestDatabase(CommandTest):

    def test_init_default(self):
        cmd = ['db', 'init']
        target = self._run(cmd, patch='alchemist.db.init')

        target.assert_called()
        target.assert_called_with(
            echo=False, names=[], verbose=True, commit=True, offline=False,
            database=False)

    def test_init_echo(self):
        cmd = ['db', 'init', '--echo']
        target = self._run(cmd, patch='alchemist.db.init')

        target.assert_called()
        target.assert_called_with(
            echo=True, names=[], verbose=True, commit=True, offline=False,
            database=False)

    def test_init_offline(self):
        cmd = ['db', 'init', '--offline', '--echo']
        target = self._run(cmd, patch='alchemist.db.init')

        target.assert_called()
        target.assert_called_with(
            echo=True, names=[], verbose=True, commit=True, offline=True,
            database=False)

    def test_clear_default(self):
        cmd = ['db', 'clear']
        target = self._run(cmd, patch='alchemist.db.clear')

        target.assert_called()
        target.assert_called_with(
            echo=False, names=[], verbose=True, commit=True, offline=False,
            database=False)

    def test_flush_default(self):
        cmd = ['db', 'flush']
        target = self._run(cmd, patch='alchemist.db.flush')

        target.assert_called()
        target.assert_called_with(
            echo=False, names=[], verbose=True, commit=True, offline=False)


class TestSettings(CommandTest):

    def test_settings_default(self):
        capture = py.io.StdCapture(out=True, in_=False)
        self._run(['settings'], patch='alchemist.db.init')
        out, err = capture.done()

        text = out.read()

        assert "'COMPONENTS':" in text
